/*
 * Sample for eDavki SOAP Service
 *
 * requires Apache Axis 1.1 and XML Security 1.1.0
 */
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.io.FileInputStream;
import java.net.URL;
import java.net.MalformedURLException;
import java.text.SimpleDateFormat;
import java.util.Date;

import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.Principal;
import java.security.PrivateKey;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import java.rmi.RemoteException;

import javax.xml.rpc.ServiceException;
import javax.xml.rpc.holders.StringHolder;
import javax.xml.transform.TransformerException;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.xpath.XPathAPI;
import org.apache.xpath.objects.XObject;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.apache.xml.utils.PrefixResolver;

import org.apache.xml.security.exceptions.XMLSecurityException;
import org.apache.xml.security.signature.XMLSignature;
import org.apache.xml.security.signature.XMLSignatureException;
import org.apache.xml.security.transforms.Transforms;
import org.apache.xml.security.utils.Constants;
import org.apache.xml.security.utils.XMLUtils;
import org.apache.xml.serialize.OutputFormat;

import si.durs.edavki.beta.WebServices.SoapPortal_1.*;

/**
 * @author mark
 *  
 */
public class TEdpSoapService {
    static String serviceURL = "https://dursweb.hermes.si/SoapPortal/EdpSoapService.asmx"; // mark
    static String documentDomain = "dursweb.hermes.si";

    static String NF_PrevZav_URI = "http://edavki.durs.si/Documents/Schemas/NF_PrevZav_1.xsd";
    static String XML_DSIG_URI = "http://www.w3.org/2000/09/xmldsig#";
    static String EDP_Common_URI = "http://edavki.durs.si/Documents/Schemas/EDP-Common-1.xsd";
    static String HslSig_URI = "urn:schemas-hermes-softlab-com:2003/09/Signatures";
    
    static int taxNumber = 20763808;
    static String taxPayerType = "FO";
    static String signer = "Marko Macek";
    
    static final String keyStoreFile = "c:/work/test1.p12";
    static final String keyStoreType = "pkcs12";
    static String keyStorePass = "test1";
    static final String keyAlias = "{6eadaf2d-5bf4-49db-9d01-91b35c8beaa1}";
    static final String keyPass = keyStorePass;
    
    static final String trustStoreFile = "c:/work/trust.jks";
    static final String trustStorePass = "changeit"; 

    public static void main(String[] args) {
        try {
            URL url = new URL(serviceURL);
            EdpSoapService service = new EdpSoapServiceLocator();
            EdpSoapServiceSoap soap = service.getEdpSoapServiceSoap(url);
            
            if (args.length > 0)
                keyStorePass = args[0];
            
            //System.setProperty("javax.net.debug", "all");
            System.setProperty("javax.net.ssl.keyStore", keyStoreFile);
            System.setProperty("javax.net.ssl.keyStoreType", keyStoreType);
            System.setProperty("javax.net.ssl.keyStorePassword", keyStorePass);
            System.setProperty("javax.net.ssl.trustStore", trustStoreFile);
            System.setProperty("javax.net.ssl.trustStorePassword", trustStorePass);
            
            org.apache.xml.security.Init.init();
            
            KeyStore ks = KeyStore.getInstance("PKCS12");
            FileInputStream fis = new FileInputStream(keyStoreFile);

            //load the keystore
            ks.load(fis, keyStorePass.toCharArray());

            //get the private key for signing.
            PrivateKey privateKey = (PrivateKey) ks.getKey(keyAlias,
                                                   keyPass.toCharArray());
            
            X509Certificate cert =
                (X509Certificate) ks.getCertificate(keyAlias);

            String handle = soap.loginUsingClientCertificate(taxNumber, taxPayerType);

            String document = soap.getNewDocument(handle, "NF_PrevZav", 1);

            document = addPayload(document);

            String errors = soap.checkDocumentErrors(handle, document);

            document = soap.updatePrefilledDocumentData(handle, document);
            
            StringHolder displayTransformURL = new StringHolder();
            StringHolder displayTransformHashSHA1 = new StringHolder();
            
            soap.getDocumentDisplayTransform(handle, "NF_PrevZav", 1, 
                    displayTransformURL, displayTransformHashSHA1);

            document = signDocument(document, cert, privateKey, 
                    displayTransformURL.value, displayTransformHashSHA1.value);

            String documentId = soap.depositDocument(handle, document, false, null);

            soap.logout(handle);
        } catch (ServiceException e) {
            System.out.println("exception: " + e);
        } catch (SAXException e) {
            System.out.println("exception: " + e);
        } catch (TransformerException e) {
            System.out.println("exception: " + e);
        } catch (ParserConfigurationException e) {
            System.out.println("exception: " + e);
        } catch (RemoteException e) {
            System.out.println("exception: " + e);
        } catch (MalformedURLException e) {
            System.out.println("exception: " + e);
        } catch (KeyStoreException e) {
            System.out.println("exception: " + e);
        } catch (CertificateException e) {
            System.out.println("exception: " + e);
        } catch (IOException e) {
            System.out.println("exception: " + e);
        } catch (NoSuchAlgorithmException e) {
            System.out.println("exception: " + e);
        } catch (UnrecoverableKeyException e) {
            System.out.println("exception: " + e);
        } catch (XMLSignatureException e) {
            System.out.println("exception: " + e);
        } catch (XMLSecurityException e) {
            System.out.println("exception: " + e);
        }
    }

    static String addPayload(String document)
        throws 
            ParserConfigurationException, IOException, SAXException,
            TransformerException 
    {
        Document doc = loadDocument(document);

        PrefixResolver ns = new XPathPrefixResolver(NF_PrevZav_URI);

        Element payload = (Element) selectSingleNode(doc.getDocumentElement(),
                "//doc:body/doc:NF_PrevZav", ns);
                
        Element verify = doc.createElementNS(NF_PrevZav_URI, "Verify");
        Element taxNumber = doc.createElementNS(NF_PrevZav_URI, "TaxNumber");
        verify.appendChild(taxNumber);
        taxNumber.appendChild(doc.createTextNode("12345679"));
        Element criteria = doc.createElementNS(NF_PrevZav_URI, "Criteria");
        verify.appendChild(criteria);
        
        payload.appendChild(verify);
        
        document = serializeDocument(doc);
        return document;
    }
    
    static Element prepareSignature(Document doc, PrefixResolver ns,
            String displayTransformURL, String displayTransformHashSHA1, X509Certificate cert) 
        throws TransformerException 
    {
        Element depositorSignature = 
            doc.createElementNS(EDP_Common_URI, "edp:DepositorSignature");
        Element signatures = (Element)selectSingleNode(doc.getDocumentElement(),
                "//edp:Signatures", ns);
        signatures.appendChild(depositorSignature);
        Element depositor = doc.createElementNS(EDP_Common_URI, "edp:Depositor");
        depositorSignature.appendChild(depositor);
        Element timestamp = doc.createElementNS(EDP_Common_URI, "edp:timestamp");
        depositor.appendChild(timestamp);
        Element name = doc.createElementNS(EDP_Common_URI, "edp:name");
        depositor.appendChild(name);
        name.appendChild(doc.createTextNode(getCNName(cert)));

        Date now = new Date();
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        String dateNow = sdf.format(now);
        timestamp.appendChild(doc.createTextNode(dateNow));
        
        Element presentation = doc.createElementNS(HslSig_URI, "hslsig:Presentation");
        depositorSignature.appendChild(presentation);
        presentation.setAttribute("xmlns:hslsig", HslSig_URI);
        
        Element displayTransform = doc.createElementNS(HslSig_URI, "hslsig:DisplayTransformDigest");
        presentation.appendChild(displayTransform);
        displayTransform.setAttribute("transform", displayTransformURL);
        displayTransform.setAttribute("digestAlgorithm", "http://www.w3.org/2000/09/xmldsig#sha1");
        displayTransform.setAttribute("digest", displayTransformHashSHA1);
        
        Element signingComponent = doc.createElementNS(HslSig_URI, "hslsig:SigningComponent");
        presentation.appendChild(signingComponent);
        signingComponent.setAttribute("name", "SoapPortalJavaSample");
        signingComponent.setAttribute("version", "0");

        Element header = (Element)selectSingleNode(doc.getDocumentElement(),
                "//edp:Header", ns);
        Element domain = (Element)selectSingleNode(header,
                "edp:domain", ns);
        if (domain == null) {
            domain = doc.createElementNS(EDP_Common_URI, "edp:domain");
            header.appendChild(domain);
        }
        while (domain.hasChildNodes())
            domain.removeChild(domain.getFirstChild());
        domain.appendChild(doc.createTextNode(documentDomain));       
        
        return depositorSignature;
    }
    
    static String getCNName(X509Certificate cert) {
    	Principal subject = cert.getSubjectDN();
    	//subject.
    	String cname = subject.getName();
    	
    	int index = cname.indexOf("CN=");
    	if (index != -1) {
    		cname = cname.substring(index + 3);
    		
    		index = cname.indexOf(',');
    		if (index != -1)
    			cname = cname.substring(0, index);
    		index = cname.indexOf('+');
    		if (index != -1)
    			cname = cname.substring(0, index);
    	}
    	
		return cname.trim();
    }

    static String signDocument(String document, 
            X509Certificate cert, PrivateKey privateKey, 
            String displayTransformURL, String displayTransformHashSHA1)
        throws IOException, ParserConfigurationException, SAXException, 
        TransformerException, XMLSignatureException, XMLSecurityException
    {
        Document doc = loadDocument(document);
        PrefixResolver ns = new XPathPrefixResolver(NF_PrevZav_URI);

        Element signature = prepareSignature(doc, ns, 
                displayTransformURL, displayTransformHashSHA1, cert);
        
        Constants.setSignatureSpecNSprefix("");
        XMLSignature sig = 
            new XMLSignature(doc, null, XMLSignature.ALGO_ID_SIGNATURE_RSA);
        sig.setId("DepositorSignature");
        signature.appendChild(sig.getElement());

        Transforms transforms = new Transforms(doc);
        transforms.addTransform(Transforms.TRANSFORM_ENVELOPED_SIGNATURE);
        sig.addDocument("", transforms, Constants.ALGO_ID_DIGEST_SHA1);
        
        sig.addKeyInfo(cert);
        sig.addKeyInfo(cert.getPublicKey());

        sig.sign(privateKey);
        
        document = serializeDocument(doc);
        return document;
    }

    static Document loadDocument(String document)
        throws ParserConfigurationException, IOException, SAXException,
            TransformerException 
    {
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        dbf.setNamespaceAware(true);
        dbf.setIgnoringComments(false);
        dbf.setIgnoringElementContentWhitespace(false);
        dbf.setValidating(false);

        DocumentBuilder docBuilder = dbf.newDocumentBuilder();

        Reader r = new StringReader(document);
        Document doc = docBuilder.parse(new InputSource(r));

        return doc;
    }

    static String serializeDocument(Document doc) 
        throws TransformerException, IOException 
    {

        ByteArrayOutputStream outb = new ByteArrayOutputStream();
        BufferedOutputStream buffOut = new BufferedOutputStream(outb);

        try {
            OutputFormat outputFormat = new OutputFormat(doc);
            outputFormat.setEncoding("UTF-8");
            XMLUtils.outputDOM(doc, buffOut, true);
            buffOut.flush();
        } finally {
            buffOut.close();
        }
        return outb.toString("UTF-8");
        //return outb.
/*    	
    	// Prepare the DOM document for writing
        Source source = new DOMSource(doc);

        // Prepare the output file
        StringWriter w = new StringWriter();
        Result result = new StreamResult(w);

        // Write the DOM document to the file
        Transformer xformer = TransformerFactory.newInstance().newTransformer();
        xformer.transform(source, result);
        return w.toString();
*/
    }

    private static Node selectSingleNode(Node node, String xpath,
            PrefixResolver ns) 
    throws TransformerException 
    {
        XObject obj = XPathAPI.eval(node, xpath, ns);
        return obj.nodeset().nextNode();
    }

    private static class XPathPrefixResolver implements PrefixResolver {
        String fDocumentURI;

        XPathPrefixResolver(String documentURI) {
            fDocumentURI = documentURI;
        }

        public boolean handlesNullPrefixes() {
            return false;
        }

        public String getBaseIdentifier() {
            return null;
        }

        public String getNamespaceForPrefix(String prefix) {
            if (prefix.equals("ds"))
                return XML_DSIG_URI;
            if (prefix.equals("edp"))
                return EDP_Common_URI;
            if (prefix.equals("doc"))
                return fDocumentURI;
            return null;
        }

        public String getNamespaceForPrefix(String prefix, Node arg1) {
            return getNamespaceForPrefix(prefix);
        }
    }
}
    
